#include "stdafx.h"
#include "ODBCWrappers.h"
#include <array>


#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//The ODBC_CHECK_RETURN2 macro checks and sets the return value from ODBC calls
#define ODBC_CHECK_RETURN2(handle) \
  if (handle) \
    handle.ValidateReturnValue(m_nRet); \
  if (!SQL_SUCCEEDED(m_nRet)) \
  { \
    m_RETURNVALUE = UnknownODBCError; \
    if (handle) \
      handle.GetDiagRecords(m_sDiagRecords); \
    return m_RETURNVALUE; \
  }

//Another flavour of an ODBC_CHECK_RETURN macro
#define ODBC_CHECK_RETURN3(nRet, handle) \
  if (handle) \
    handle.ValidateReturnValue(nRet); \
  if (!SQL_SUCCEEDED(nRet)) \
  { \
    return nRet; \
  }


//test ODBC accessor class used for demonstrating calling a stored procedure
class CdboPJTestAccessor //NOLINT(clang-analyzer-optin.performance.Padding)
{
public:
//Enums
  enum SPError
  {
    UnknownODBCError = -1000,
    NoRowsetReturned = -1002,
    NotRun = -1004,
  };

//Parameter values
  LONG m_RETURNVALUE;
  LONG m_nValue;
  bool m_bValue;
  TCHAR m_nChar;
  BYTE m_nX;
  SQLLEN m_nIndicatorX;
  float m_fReal;
  double m_fFloat;
  BYTE m_byTinyInt;
  short m_nSmallInt;
  __int64 m_nBigInt;
  BYTE m_byBinary[1024]; //NOLINT(modernize-avoid-c-arrays)
  SQLINTEGER m_nBinaryLength;
  SQLLEN m_nBinaryStatus;
  SQL_NUMERIC_STRUCT m_Decimal;
  SQLLEN m_DecimalStatus;
  TCHAR m_sDecimal[64]; //NOLINT(modernize-avoid-c-arrays)
  BYTE m_byBinaryOut[1024]; //NOLINT(modernize-avoid-c-arrays)
  SQLINTEGER m_nBinaryOutLength;
  SQLLEN m_nBinaryOutStatus;
  SQL_NUMERIC_STRUCT m_DecimalOut;
  TCHAR m_sDecimalOut[64]; //NOLINT(modernize-avoid-c-arrays)
  SQLLEN m_nDecimalOutStatus;
  double m_fDecimalOut;

//Column values
  LONG  m_colfldAddressID;
  TCHAR m_colfldAddressLine1[61]; //NOLINT(modernize-avoid-c-arrays)
  TCHAR m_colfldAddressLine2[61]; //NOLINT(modernize-avoid-c-arrays)
  TCHAR m_colfldCity[31]; //NOLINT(modernize-avoid-c-arrays)
  LONG  m_colfldStateProvinceID;
  TCHAR m_colfldPostalCode[16]; //NOLINT(modernize-avoid-c-arrays)
  GUID  m_colfldrowguid;
  TCHAR  m_colfldChar;
  TIMESTAMP_STRUCT m_colfldModifiedDate;

BEGIN_ODBC_PARAM_MAP(CdboPJTestAccessor)
  SET_ODBC_PARAM_TYPE(SQL_PARAM_OUTPUT)
  ODBC_PARAM_ENTRY(1, m_RETURNVALUE)
  SET_ODBC_PARAM_TYPE(SQL_PARAM_INPUT)
  ODBC_PARAM_ENTRY(2, m_nValue)
  ODBC_PARAM_ENTRY(3, m_bValue)
  ODBC_PARAM_ENTRY(4, m_nChar)
  ODBC_PARAM_ENTRY(5, m_fReal)
  ODBC_PARAM_ENTRY(6, m_fFloat)
  ODBC_PARAM_ENTRY(7, m_byTinyInt)
  ODBC_PARAM_ENTRY(8, m_nSmallInt)
  ODBC_PARAM_ENTRY(9, m_nBigInt)
  ODBC_PARAM_LENGTH_STATUS(10, m_byBinary, m_nBinaryLength, m_nBinaryStatus)
  ODBC_PARAM_PS_LENGTH_STATUS(11, m_sDecimal, 2, 18, m_DecimalStatus)
  SET_ODBC_PARAM_TYPE(SQL_PARAM_INPUT_OUTPUT)
  ODBC_PARAM_ENTRY_STATUS(12, m_nX, m_nIndicatorX)
  ODBC_PARAM_LENGTH_STATUS(13, m_byBinaryOut, m_nBinaryOutLength, m_nBinaryOutStatus)
  ODBC_PARAM_ENTRY_STATUS(14, m_DecimalOut, m_nDecimalOutStatus)
END_ODBC_PARAM_MAP()

BEGIN_ODBC_COLUMN_MAP(CdboPJTestAccessor)
  ODBC_COLUMN_ENTRY(1, m_colfldAddressID)
  ODBC_COLUMN_ENTRY(2, m_colfldAddressLine1)
  ODBC_COLUMN_ENTRY(3, m_colfldAddressLine2)
  ODBC_COLUMN_ENTRY(4, m_colfldCity)
  ODBC_COLUMN_ENTRY(5, m_colfldStateProvinceID)
  ODBC_COLUMN_ENTRY(6, m_colfldPostalCode)
  ODBC_COLUMN_ENTRY(7, m_colfldrowguid)
  ODBC_COLUMN_ENTRY(8, m_colfldModifiedDate)
  ODBC_COLUMN_ENTRY(9, m_colfldChar)
END_ODBC_COLUMN_MAP()

DEFINE_ODBC_COMMAND(CdboPJTestAccessor, _T("{ ? = CALL dbo.PJ_Test(?,?,?,?,?,?,?,?,?,?,?,?,?) }"))

  //You may wish to call this function if you are inserting a record and wish to
  //initialize all the fields, if you are not going to explicitly set all of them.
  void ClearRecord() noexcept
  {
    memset(this, 0, sizeof(*this));
  }
};


//Another test accessor class
class CdboPJTest2Accessor
{
public:
  TCHAR m_szTableName[128]; //NOLINT(modernize-avoid-c-arrays)
};


//test class used for demonstrating calling a stored procedure
class CdboPJTest : public CODBC::CAccessor<CdboPJTestAccessor>
{
public:
//Constructors / Destructors
  explicit CdboPJTest(CODBC::CConnection* pDbConnect) noexcept : m_nRet{SQL_SUCCESS},
                                                                 m_pDbConnect{pDbConnect}
  {
#pragma warning(suppress: 26477)
    ATLASSUME(m_pDbConnect != nullptr);

    ClearRecord();
    m_RETURNVALUE = NotRun;
  }

//Methods
  int RunStoredProcedure()
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(m_pDbConnect != nullptr);

    //Create the statement object
    CODBC::CStatement statement;
    m_nRet = statement.Create(*m_pDbConnect);
    ODBC_CHECK_RETURN2(statement);

#pragma warning(suppress: 26465 26492)
    m_nRet = statement.ColumnPrivileges(nullptr, nullptr, const_cast<TCHAR*>(_T("Person.Address")), nullptr);
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.Columns(nullptr, const_cast<TCHAR*>(_T("Person")), const_cast<TCHAR*>(_T("Address")), const_cast<TCHAR*>(_T("AddressLine1")));
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.Tables(nullptr, const_cast<TCHAR*>(_T("Person")), const_cast<TCHAR*>(_T("Address")), nullptr);
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.Procedures(nullptr, const_cast<TCHAR*>(_T("dbo")), const_cast<TCHAR*>(_T("PJ_Test")));
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.ProcedureColumns(nullptr, const_cast<TCHAR*>(_T("dbo")), const_cast<TCHAR*>(_T("PJ_Test")), nullptr);
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.PrimaryKeys(nullptr, const_cast<TCHAR*>(_T("Person")), const_cast<TCHAR*>(_T("Address")));
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.ForeignKeys(nullptr, const_cast<TCHAR*>(_T("State")), const_cast<TCHAR*>(_T("Province")), nullptr, const_cast<TCHAR*>(_T("Person")), const_cast<TCHAR*>(_T("Address")));
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.TablePrivileges(nullptr, const_cast<TCHAR*>(_T("Person")), nullptr);
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.Statistics(nullptr, const_cast<TCHAR*>(_T("Person")), const_cast<TCHAR*>(_T("Address")), SQL_INDEX_ALL, SQL_ENSURE);
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
#pragma warning(suppress: 26465 26492)
    m_nRet = statement.SpecialColumns(SQL_BEST_ROWID, nullptr, const_cast<TCHAR*>(_T("Person")), const_cast<TCHAR*>(_T("Address")), SQL_SCOPE_CURROW, SQL_NULLABLE);
    ODBC_CHECK_RETURN2(statement);
    statement.Cancel();
    m_nRet = statement.GetTypeInfo(SQL_DECIMAL);
    ODBC_CHECK_RETURN2(statement);

    CODBC::CDynamicColumnAccessor<CdboPJTest2Accessor> accessor;
    accessor.BindColumns(statement);

    //Iterate through the returned recordset
    bool bContinue{true};
    while (bContinue)
    {
      m_nRet = statement.FetchNext();
      if (!SQL_SUCCEEDED(m_nRet))
        bContinue = false;
      else
      {
        const SQLSMALLINT nColumn1{accessor.GetColumnNo(_T("TYPE_NAME"))};
        UNREFERENCED_PARAMETER(nColumn1);
        const SQLSMALLINT nColumn2{accessor.GetColumnNo(_T("COLUMN_NAME"))};
        UNREFERENCED_PARAMETER(nColumn2);
#pragma warning(suppress: 26485)
        const SQLRETURN nRet{accessor.GetValue(statement, 1, accessor.m_szTableName)};
        ODBC_CHECK_RETURN(nRet, statement);
      }
    }
    statement.Free(SQL_CLOSE);

    SQLUINTEGER nAttributes{0};
    m_nRet = m_pDbConnect->GetInfo(SQL_DYNAMIC_CURSOR_ATTRIBUTES1, &nAttributes, sizeof(nAttributes), nullptr);
    ODBC_CHECK_RETURN2(statement);
    m_nRet = statement.SetAttr(SQL_ATTR_CURSOR_TYPE, SQL_CURSOR_DYNAMIC);
    ODBC_CHECK_RETURN2(statement);
    m_nRet = statement.SetAttr(SQL_ATTR_CURSOR_SCROLLABLE, SQL_SCROLLABLE);
    ODBC_CHECK_RETURN2(statement);
    m_nRet = statement.SetAttr(SQL_ATTR_CONCURRENCY, SQL_CONCUR_ROWVER);
    ODBC_CHECK_RETURN2(statement);
    m_nRet = statement.SetAttr(SQL_ATTR_CURSOR_SENSITIVITY, SQL_INSENSITIVE);
    ODBC_CHECK_RETURN2(statement);

    //Fill in the member variables which we will be binding to
    m_RETURNVALUE = 0;
    m_nValue = 33;
    m_bValue = true;
    m_nChar = 12;
    m_nX = 55;
    m_nIndicatorX = 0;
    m_fReal = static_cast<float>(12.3);
    m_fFloat = 15.7;
    m_byTinyInt = 33;
    m_nSmallInt = 44;
    m_nBigInt = 77;
#pragma warning(suppress: 26485)
    memset(m_byBinary, 0x31, sizeof(m_byBinary));
    m_nBinaryLength = 4;
    m_nBinaryStatus = 4;
    m_nBinaryOutLength = sizeof(m_byBinaryOut);
    
    m_Decimal.precision = 18;
    m_Decimal.scale = 2;
    m_Decimal.sign = 1;
#pragma warning(suppress: 26485)
    memset(m_Decimal.val, 0, sizeof(m_Decimal.val));
    m_DecimalStatus = SQL_NTS;
#pragma warning(suppress: 26485)
    _tcscpy_s(m_sDecimal, _countof(m_sDecimal), _T("123.45"));
    m_nDecimalOutStatus = SQL_NULL_DATA;

    //Prepare the statement
    m_nRet = statement.Prepare(GetDefaultCommand());
    ODBC_CHECK_RETURN2(statement);

    //Bind the parameters
    m_nRet = BindParameters(statement);
    ODBC_CHECK_RETURN2(statement);

    //Execute the statement
    m_nRet = statement.Execute();
    ODBC_CHECK_RETURN2(statement);

    //Check the return value
    if (m_RETURNVALUE != 0)
      return m_RETURNVALUE;

    //Bind the columns
    m_nRet = BindColumns(statement);
    ODBC_CHECK_RETURN2(statement);

    SQLHDESC hArd0{nullptr};
    SQLHDESC hIrd0{nullptr};
    m_nRet = statement.GetAttr(SQL_ATTR_APP_ROW_DESC, &hArd0, 0, nullptr);
    ODBC_CHECK_RETURN2(statement);
    m_nRet = statement.GetAttr(SQL_ATTR_IMP_ROW_DESC, &hIrd0, 0, nullptr);
    ODBC_CHECK_RETURN2(statement);
    CODBC::CDescriptor descriptor1;
    descriptor1.Attach(hArd0);
    CODBC::CDescriptor descriptor2;
    descriptor2.Attach(hIrd0);

    CODBC::CDescriptor descriptor3;
    m_nRet = descriptor3.Create(*m_pDbConnect);
    descriptor3.ValidateReturnValue(m_nRet);
    m_nRet = descriptor1.Copy(descriptor3);
    descriptor1.ValidateReturnValue(m_nRet);
    SQLSMALLINT nField{0};
    SQLINTEGER nSize{0};
    m_nRet = descriptor3.GetField(0, SQL_DESC_ALLOC_TYPE, &nField, sizeof(nField), &nSize);
    descriptor3.ValidateReturnValue(m_nRet);
    ATL::CAtlString sName;
    SQLSMALLINT nStringLength{0};
    SQLSMALLINT nType{0};
    SQLSMALLINT nSubType{0};
    SQLLEN nLength{0};
    SQLSMALLINT nPrecision{0};
    SQLSMALLINT nScale{0};
    SQLSMALLINT nNullable{0};
#pragma warning(suppress: 26473 26490)
    m_nRet = descriptor3.GetRec(1, reinterpret_cast<SQLTCHAR*>(sName.GetBuffer(128)), 128, &nStringLength, &nType, &nSubType, &nLength, &nPrecision, &nScale, &nNullable);
    sName.ReleaseBuffer();
    descriptor3.ValidateReturnValue(m_nRet);
    m_nRet = descriptor3.SetRec(1, nType, nSubType, nLength, nPrecision, nScale, nullptr, nullptr, nullptr);
    descriptor3.ValidateReturnValue(m_nRet);
    descriptor3.SetField(0, SQL_DESC_ALLOC_TYPE, &nField, sizeof(nField));
    descriptor3.ValidateReturnValue(m_nRet);

    //Check we have got back a results set
    SQLSMALLINT nColumns{0};
    m_nRet = statement.NumResultCols(&nColumns);
    ODBC_CHECK_RETURN2(statement);
    if (nColumns == 0)
    {
      m_RETURNVALUE = NoRowsetReturned;
      return m_RETURNVALUE;
    }

  #ifdef _DEBUG
    SQLSMALLINT nNumParams{0};
    const SQLRETURN nRet{statement.NumParams(&nNumParams)};
    UNREFERENCED_PARAMETER(nRet);
    SQLLEN nTemp{0};
    m_nRet = statement.ColAttribute(3, SQL_DESC_UNNAMED, nullptr, SQL_IS_POINTER, nullptr, &nTemp);
    ODBC_CHECK_RETURN2(statement);
  #endif //#ifdef _DEBUG

    //Iterate through the returned recordset
    bContinue = true;
    while (bContinue)
    {
      ClearRecord();
      m_nRet = statement.FetchNext();
      if (!SQL_SUCCEEDED(m_nRet))
        bContinue = false;
      else
      {
        //m_nRet = GetData(statement); //Used if using a dynamic accessor
        ODBC_CHECK_RETURN2(statement);
      }
    }

    return m_RETURNVALUE;
  }

  /* //Used if using a dynamic accessor
  SQLRETURN GetData(CODBC::CStatement& statement)
  {
    SQLINTEGER nColumn{GetColumnNo(_T("AddressID"))};
    nColumn = GetColumnNo(_T("AddressLine1"));
    nColumn = GetColumnNo(_T("AddressLine2"));
    nColumn = GetColumnNo(_T("City"));
    nColumn = GetColumnNo(_T("StateProvinceID"));
    nColumn = GetColumnNo(_T("PostalCode"));
    nColumn = GetColumnNo(_T("rowguid"));
    nColumn = GetColumnNo(_T("ModifiedDate"));

    SQLRETURN nRet{GetValue(statement, 1, m_colfldAddressID)};
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 2, m_colfldAddressLine1);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 3, m_colfldAddressLine2);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 4, m_colfldCity);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 5, m_colfldStateProvinceID);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 6, m_colfldPostalCode);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 7, m_colfldrowguid);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 8, m_colfldModifiedDate);
    ODBC_CHECK_RETURN(nRet, statement);
    nRet = GetValue(statement, 9, m_colfldChar);
    return nRet;
  }
  */

//Member variables
  SQLRETURN m_nRet;
  CODBC::CConnection* m_pDbConnect;
  std::vector<CODBC::String> m_sDiagRecords;
};


//test ODBC accessor class used for demonstrating inserting a row into AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryInsertAccessor
{
public:
//Parameter values
  TCHAR m_sProductCategory[51]; //NOLINT(modernize-avoid-c-arrays)

BEGIN_ODBC_PARAM_MAP(CdboProductCategoryInsertAccessor)
  SET_ODBC_PARAM_TYPE(SQL_PARAM_INPUT)
  ODBC_PARAM_ENTRY(1, m_sProductCategory)
END_ODBC_PARAM_MAP()

DEFINE_ODBC_COMMAND(CdboProductCategoryInsertAccessor, _T("INSERT INTO Production.ProductCategory (Name) VALUES (?)"))

  //You may wish to call this function if you are inserting a record and wish to
  //initialize all the fields, if you are not going to explicitly set all of them.
  void ClearRecord() noexcept
  {
    memset(this, 0, sizeof(*this));
  }
};


//test class used for demonstrating inserting a row into  AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryInsert : public CODBC::CAccessor<CdboProductCategoryInsertAccessor>
{
public:
//Methods
  SQLRETURN Insert(const CODBC::CConnection* pDbConnect)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(pDbConnect != nullptr);

    //Create the statement object
    CODBC::CStatement statement;
    SQLRETURN nRet = statement.Create(*pDbConnect);
    ODBC_CHECK_RETURN3(nRet, statement);

    CODBC::String sTemp;
    nRet = statement.GetCursorName(sTemp); //NOLINT(clang-analyzer-deadcode.DeadStores)

    //Prepare the statement
    nRet = statement.Prepare(GetDefaultCommand());
    ODBC_CHECK_RETURN3(nRet, statement);

    //Bind the parameters
#pragma warning(suppress: 26485)
    _tcscpy_s(m_sProductCategory, _countof(m_sProductCategory), _T("Unicycles"));
    nRet = BindParameters(statement);
    ODBC_CHECK_RETURN3(nRet, statement);

    //Execute the statement
    nRet = statement.Execute();
    ODBC_CHECK_RETURN3(nRet, statement);
    return nRet;
  }
};


//test ODBC accessor class used for demonstrating returning rows from AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryIterateAccessor
{
public:
//Parameter values
  LONG m_colfldProductCategoryID;
  TCHAR m_colfldProductCategory[51]; //NOLINT(modernize-avoid-c-arrays)
  GUID m_colfldrowguid;
  SQL_TIMESTAMP_STRUCT m_colfldModifiedDate;

BEGIN_ODBC_PARAM_MAP(CdboProductCategoryIterateAccessor)
END_ODBC_PARAM_MAP()

BEGIN_ODBC_COLUMN_MAP(CdboProductCategoryIterateAccessor)
  ODBC_COLUMN_ENTRY(1, m_colfldProductCategoryID)
  ODBC_COLUMN_ENTRY(2, m_colfldProductCategory)
  ODBC_COLUMN_ENTRY(3, m_colfldrowguid)
  ODBC_COLUMN_ENTRY(4, m_colfldModifiedDate)
END_ODBC_COLUMN_MAP()

DEFINE_ODBC_COMMAND(CdboProductCategoryIterateAccessor, _T("SELECT ProductCategoryID, Name, rowguid, ModifiedDate FROM Production.ProductCategory"))

  //You may wish to call this function if you are inserting a record and wish to
  //initialize all the fields, if you are not going to explicitly set all of them.
  void ClearRecord() noexcept
  {
    memset(this, 0, sizeof(*this));
  }
};


//test class used for demonstrating iterating across the AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryIterate : public CODBC::CCommand<CODBC::CAccessor<CdboProductCategoryIterateAccessor>>
{
public:
//Methods
  SQLRETURN Iterate(const CODBC::CConnection* pDbConnect, _In_ bool bBind = true, _In_opt_ CODBC::SQL_ATTRIBUTE* pAttributes = nullptr, _In_ ULONG nAttributes = 0)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(pDbConnect != nullptr);

    SQLRETURN nRet{Open(*pDbConnect, GetDefaultCommand(), bBind, pAttributes, nAttributes)};
    ODBC_CHECK_RETURN3(nRet, m_Command);

    //Iterate through the returned recordset
    bool bContinue{true};
    while (bContinue)
    {
      ClearRecord();
      nRet = m_Command.FetchNext();
      if (!SQL_SUCCEEDED(nRet))
        bContinue = false;
    }

    return SQL_SUCCESS;
  }
};


//test ODBC accessor class used for demonstrating updating a row in AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryUpdateAccessor
{
public:
//Parameter values
  TCHAR m_sProductCategory[51]; //NOLINT(modernize-avoid-c-arrays)

BEGIN_ODBC_PARAM_MAP(CdboProductCategoryUpdateAccessor)
  SET_ODBC_PARAM_TYPE(SQL_PARAM_INPUT)
  ODBC_PARAM_ENTRY(1, m_sProductCategory)
END_ODBC_PARAM_MAP()

DEFINE_ODBC_COMMAND(CdboProductCategoryInsertAccessor, _T("UPDATE Production.ProductCategory SET Name = ? WHERE Name='Unicycles'"))

  //You may wish to call this function if you are inserting a record and wish to
  //initialize all the fields, if you are not going to explicitly set all of them.
  void ClearRecord() noexcept
  {
    memset(this, 0, sizeof(*this));
  }
};


//test CCommand class used for demonstrating updating a row in AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryUpdate : public CODBC::CCommand<CODBC::CAccessor<CdboProductCategoryUpdateAccessor>>
{
public:
//Methods
  SQLRETURN Update(const CODBC::CConnection* pDbConnect)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(pDbConnect != nullptr);

    //Bind the parameters
#pragma warning(suppress: 26485)
    _tcscpy_s(m_sProductCategory, _countof(m_sProductCategory), _T("Tricycles"));

    //Execute the statement
    const SQLRETURN nRet{Open(*pDbConnect, GetDefaultCommand())};
    ODBC_CHECK_RETURN3(nRet, m_Command);

    return nRet;
  }
};


//test class used for demonstrating deleting a row from the AdventureWorks2008.Production.ProductCategory table
class CdboProductCategoryDelete
{
public:
//Methods
  SQLRETURN Delete(const CODBC::CConnection* pDbConnect)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSUME(pDbConnect != nullptr);

    //Create the statement object
    CODBC::CStatement statement;
    SQLRETURN nRet = statement.Create(*pDbConnect);
    ODBC_CHECK_RETURN3(nRet, statement);

    //Prepare the statement
#pragma warning(suppress: 26465 26490 26492)
    nRet = statement.Prepare(const_cast<SQLTCHAR*>(reinterpret_cast<const SQLTCHAR*>(_T("DELETE FROM Production.ProductCategory WHERE Name = 'Tricycles'"))));
    ODBC_CHECK_RETURN3(nRet, statement);

    //Execute the statement
    nRet = statement.Execute();
    ODBC_CHECK_RETURN3(nRet, statement);
    return nRet;
  }
};


int _tmain()
{
  //Setup the environment
  CODBC::CEnvironment env;
  SQLRETURN nRet{env.Create()};
  env.ValidateReturnValue(nRet);
  nRet = env.SetAttr(SQL_ATTR_ODBC_VERSION, SQL_OV_ODBC3);
  env.ValidateReturnValue(nRet);

  nRet = env.SetAttrU(SQL_ATTR_CONNECTION_POOLING, SQL_CP_OFF);
  env.ValidateReturnValue(nRet);

  SQLINTEGER nVersion;
  nRet = env.GetAttr(SQL_ATTR_ODBC_VERSION, nVersion);
  env.ValidateReturnValue(nRet);

  SQLUINTEGER nPooling;
  nRet = env.GetAttrU(SQL_ATTR_CONNECTION_POOLING, nPooling);
  env.ValidateReturnValue(nRet);

  nRet = env.CommitTran();
  env.ValidateReturnValue(nRet);
  nRet = env.RollbackTran();
  env.ValidateReturnValue(nRet);


  //Setup the connection
  CODBC::CConnection con;
  nRet = con.Create(env);
  con.ValidateReturnValue(nRet);
  std::vector<CODBC::String> records;
  con.GetDiagRecords(records);


  CODBC::String sDesc;
  std::vector<CODBC::String> sAttributes;
  nRet = env.Drivers(SQL_FETCH_FIRST, sDesc, sAttributes); //NOLINT(clang-analyzer-deadcode.DeadStores)

  CODBC::String sServer;
  nRet = env.DataSources(SQL_FETCH_FIRST, sServer, sDesc);
  while (SQL_SUCCEEDED(nRet))
  {
    ATLTRACE(_T("Server:%s\n"), sServer.c_str());
    nRet = env.DataSources(SQL_FETCH_NEXT, sServer, sDesc);
  }


  SQLHANDLE h{con.Detach()};
  con.Attach(h);
  CODBC::CConnection h3;
  h3.Attach(con.Detach());
  con.Attach(h3.Detach());

  //Set login timeout to 5 seconds
  nRet = con.SetAttrU(SQL_ATTR_LOGIN_TIMEOUT, 5);
  con.ValidateReturnValue(nRet);


  //Connect to the datasource (Assumes AdventureWorks2008 sample database is installed and made available as a DSN called "AdventureWorks")
  CODBC::String sFullInitString;
#pragma warning(suppress: 26465 26490 26492)
  nRet = con.DriverConnect(const_cast<SQLTCHAR*>(reinterpret_cast<const SQLTCHAR*>(_T("DSN=AdventureWorks;"))), sFullInitString);
  CODBC::String sTemp;
#pragma warning(suppress: 26465)
  //nRet = con.BrowseConnect(const_cast<TCHAR*>(_T("DSN=AdventureWorks")), sTemp);
  //nRet = con.Connect(_T("SQLLOCAL"), nullptr, nullptr);
  con.ValidateReturnValue(nRet);

  nRet = con.GetAttr(SQL_ATTR_CURRENT_CATALOG, sTemp);
  con.ValidateReturnValue(nRet);

  SQLUINTEGER nTemp{0};
  nRet = con.GetAttrU(SQL_ATTR_PACKET_SIZE, nTemp);
  con.ValidateReturnValue(nRet);

  SQLINTEGER nTemp2{0};
  nRet = con.GetAttr(SQL_ATTR_TRANSLATE_OPTION, nTemp2);
  con.ValidateReturnValue(nRet);

#pragma warning(suppress: 26465 26492)
  nRet = con.NativeSql(const_cast<TCHAR*>(_T("SELECT { fn CONVERT (empid, SQL_SMALLINT) } FROM employee")), sTemp);
  con.ValidateReturnValue(nRet);

  //nRet = con.SetAttr(SQL_ATTR_CURRENT_CATALOG, _T("AdventureWorks"));
  //con.ValidateReturnValue(nRet);

  nRet = con.GetAttr(SQL_ATTR_CURRENT_CATALOG, sTemp);
  con.ValidateReturnValue(nRet);

  nRet = con.CommitTran();
  con.ValidateReturnValue(nRet);
  nRet = con.RollbackTran();
  con.ValidateReturnValue(nRet);

  CODBC::CDescriptor descriptor;
  nRet = descriptor.Create(con);
  if (descriptor)
    descriptor.ValidateReturnValue(nRet);

  SQLUSMALLINT nFunctions{0};
  nRet = con.GetFunctions(SQL_API_SQLTABLES, &nFunctions);
  con.ValidateReturnValue(nRet);


  //Run our test stored procedure (Can be created using the "Create PJ_Test SP.sql" script)
  CdboPJTest sp{&con};
  const int nReturn{sp.RunStoredProcedure()};
  UNREFERENCED_PARAMETER(nReturn);

  //Test out adding a row to AdventureWorks2008.Production.ProductCategory table
  CdboProductCategoryInsert table;
  table.Insert(&con);

  //Test out iterating across all rows in the AdventureWorks2008.Production.ProductCategory table
  CdboProductCategoryIterate table2;
  std::array<CODBC::SQL_ATTRIBUTE, 2> attributes
  { {
#pragma warning(suppress: 26490)
    { SQL_ATTR_CONCURRENCY,        reinterpret_cast<SQLPOINTER>(SQL_CONCUR_ROWVER), SQL_IS_INTEGER },
#pragma warning(suppress: 26490)
    { SQL_ATTR_CURSOR_SENSITIVITY, reinterpret_cast<SQLPOINTER>(SQL_INSENSITIVE),   SQL_IS_INTEGER }
  } };
#pragma warning(suppress: 26472)
  nRet = table2.Iterate(&con, true, attributes.data(), static_cast<ULONG>(attributes.size())); //NOLINT(clang-analyzer-deadcode.DeadStores)

  //Test out updating a row in the AdventureWorks2008.Production.ProductCategory table
  CdboProductCategoryUpdate table3;
  nRet = table3.Update(&con); //NOLINT(clang-analyzer-deadcode.DeadStores)

  //Test deleting a row from the AdventureWorks2008.Production.ProductCategory table
  CdboProductCategoryDelete table4;
  nRet = table4.Delete(&con); //NOLINT(clang-analyzer-deadcode.DeadStores)

  return 0;
}
